@@ -144,11 +144,13 @@ static eap_tls_t *eap_tls_create(identification_t *server,
 	);
 
 	frag_size = lib->settings->get_int(lib->settings,
-					"charon.plugins.eap-tls.fragment_size", MAX_FRAGMENT_LEN);
+					"%s.plugins.eap-tls.fragment_size", MAX_FRAGMENT_LEN,
+					charon->name);
 	max_msg_count = lib->settings->get_int(lib->settings,
-					"charon.plugins.eap-tls.max_message_count", MAX_MESSAGE_COUNT);
+					"%s.plugins.eap-tls.max_message_count", MAX_MESSAGE_COUNT,
+					charon->name);
 	include_length = lib->settings->get_bool(lib->settings,
-					"charon.plugins.eap-tls.include_length", TRUE);
+					"%s.plugins.eap-tls.include_length", TRUE, charon->name);
 	tls = tls_create(is_server, server, peer, TLS_PURPOSE_EAP_TLS, NULL, NULL);
 	this->tls_eap = tls_eap_create(EAP_TLS, tls, frag_size, max_msg_count,
 												 include_length);
