@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2011 Andreas Steffen, HSR Hochschule fuer Technik Rapperswil
+ * Copyright (C) 2011-2012 Andreas Steffen
+ * HSR Hochschule fuer Technik Rapperswil
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -15,17 +16,16 @@
 #include "imc_scanner_state.h"
 
 #include <imc/imc_agent.h>
-#include <pa_tnc/pa_tnc_msg.h>
+#include <imc/imc_msg.h>
 #include <ietf/ietf_attr.h>
-#include <ietf/ietf_attr_pa_tnc_error.h>
+#include <ietf/ietf_attr_attr_request.h>
 #include <ietf/ietf_attr_port_filter.h>
 
-#include <tncif_names.h>
 #include <tncif_pa_subtypes.h>
 
 #include <pen/pen.h>
 #include <utils/lexparser.h>
-#include <debug.h>
+#include <utils/debug.h>
 
 #include <stdio.h>
 
@@ -33,11 +33,12 @@
 
 static const char imc_name[] = "Scanner";
 
-#define IMC_VENDOR_ID	PEN_ITA
-#define IMC_SUBTYPE		PA_SUBTYPE_ITA_SCANNER
+static pen_type_t msg_types[] = {
+	{ PEN_IETF, PA_SUBTYPE_IETF_VPN }
+};
 
 static imc_agent_t *imc_scanner;
- 
+
 /**
  * see section 3.8.1 of TCG TNC IF-IMC Specification 1.3
  */
@@ -51,8 +52,8 @@ TNC_Result TNC_IMC_Initialize(TNC_IMCID imc_id,
 		DBG1(DBG_IMC, "IMC \"%s\" has already been initialized", imc_name);
 		return TNC_RESULT_ALREADY_INITIALIZED;
 	}
-	imc_scanner = imc_agent_create(imc_name, IMC_VENDOR_ID, IMC_SUBTYPE,
-								imc_id, actual_version);
+	imc_scanner = imc_agent_create(imc_name, msg_types, countof(msg_types),
+								   imc_id, actual_version);
 	if (!imc_scanner)
 	{
 		return TNC_RESULT_FATAL;
@@ -84,6 +85,15 @@ TNC_Result TNC_IMC_NotifyConnectionChange(TNC_IMCID imc_id,
 		case TNC_CONNECTION_STATE_CREATE:
 			state = imc_scanner_state_create(connection_id);
 			return imc_scanner->create_state(imc_scanner, state);
+		case TNC_CONNECTION_STATE_HANDSHAKE:
+			if (imc_scanner->change_state(imc_scanner, connection_id, new_state,
+				&state) != TNC_RESULT_SUCCESS)
+			{
+				return TNC_RESULT_FATAL;
+			}
+			state->set_result(state, imc_id,
+							  TNC_IMV_EVALUATION_RESULT_DONT_KNOW);
+			return TNC_RESULT_SUCCESS;
 		case TNC_CONNECTION_STATE_DELETE:
 			return imc_scanner->delete_state(imc_scanner, connection_id);
 		default:
@@ -102,6 +112,7 @@ static bool do_netstat(ietf_attr_port_filter_t *attr)
 	chunk_t line, token;
 	int n = 0;
 	bool success = FALSE;
+	const char system_v4[]   = "127.0.1.1";
 	const char loopback_v4[] = "127.0.0.1";
 	const char loopback_v6[] = "::1";
 
@@ -109,12 +120,12 @@ static bool do_netstat(ietf_attr_port_filter_t *attr)
 	file = popen("/bin/netstat -n -l -p -4 -6 --inet", "r");
 	if (!file)
 	{
-		DBG1(DBG_IMC, "Failed to run netstat command");
+		DBG1(DBG_IMC, "failed to run netstat command");
 		return FALSE;
 	}
 
 	/* Read the output a line at a time */
-	while (fgets(buf, BUF_LEN-1, file))
+	while (fgets(buf, sizeof(buf), file))
 	{
 		u_char *pos;
 		u_int8_t new_protocol, protocol;
@@ -123,7 +134,7 @@ static bool do_netstat(ietf_attr_port_filter_t *attr)
 		enumerator_t *enumerator;
 		bool allowed, found = FALSE;
 
-		DBG2(DBG_IMC, "%.*s", strlen(buf)-1, buf);
+		DBG2(DBG_IMC, "%.*s", (int)(strlen(buf)-1), buf);
 
 		if (n++ < 2)
 		{
@@ -135,7 +146,7 @@ static bool do_netstat(ietf_attr_port_filter_t *attr)
 		/* Extract the IP protocol type */
 		if (!extract_token(&token, ' ', &line))
 		{
-			DBG1(DBG_IMC, "Protocol field in netstat output not found");
+			DBG1(DBG_IMC, "protocol field in netstat output not found");
 			goto end;
 		}
 		if (match("tcp", &token) || match("tcp6", &token))
@@ -148,7 +159,7 @@ static bool do_netstat(ietf_attr_port_filter_t *attr)
 		}
 		else
 		{
-			DBG1(DBG_IMC, "Skipped unknown IP protocol in netstat output");
+			DBG1(DBG_IMC, "skipped unknown IP protocol in netstat output");
 			continue;
 		}
 
@@ -163,7 +174,7 @@ static bool do_netstat(ietf_attr_port_filter_t *attr)
 		}
 		if (token.len == 0)
 		{
-			DBG1(DBG_IMC, "Local Address field in netstat output not found");
+			DBG1(DBG_IMC, "local address field in netstat output not found");
 			goto end;
 		}
 
@@ -172,13 +183,16 @@ static bool do_netstat(ietf_attr_port_filter_t *attr)
 		while (*--pos != ':' && --token.len);
 		if (*pos != ':')
 		{
-			DBG1(DBG_IMC, "Local port field in netstat output not found");
+			DBG1(DBG_IMC, "local port field in netstat output not found");
 			goto end;
 		}
 		token.len--;
 
-		/* ignore ports of IPv4 and IPv6 loopback interfaces */
-		if ((token.len == strlen(loopback_v4) &&
+		/* ignore ports of IPv4 and IPv6 loopback interfaces
+		   and the internal system IPv4 address */
+		if ((token.len == strlen(system_v4) &&
+				memeq(system_v4, token.ptr, token.len)) ||
+			(token.len == strlen(loopback_v4) &&
 				memeq(loopback_v4, token.ptr, token.len)) ||
 			(token.len == strlen(loopback_v6) &&
 				memeq(loopback_v6, token.ptr, token.len)))
@@ -199,7 +213,7 @@ static bool do_netstat(ietf_attr_port_filter_t *attr)
 			}
 		}
 		enumerator->destroy(enumerator);
-		
+
 		/* Skip the duplicate port entry */
 		if (found)
 		{
@@ -219,12 +233,13 @@ end:
 	return success;
 }
 
-static TNC_Result send_message(TNC_ConnectionID connection_id)
+/**
+ * Add IETF Port Filter attribute to the send queue
+ */
+static TNC_Result add_port_filter(imc_msg_t *msg)
 {
-	pa_tnc_msg_t *msg;
 	pa_tnc_attr_t *attr;
 	ietf_attr_port_filter_t *attr_port_filter;
-	TNC_Result result;
 
 	attr = ietf_attr_port_filter_create();
 	attr->set_noskip_flag(attr, TRUE);
@@ -234,14 +249,9 @@ static TNC_Result send_message(TNC_ConnectionID connection_id)
 		attr->destroy(attr);
 		return TNC_RESULT_FATAL;
 	}
-	msg = pa_tnc_msg_create();
 	msg->add_attribute(msg, attr);
-	msg->build(msg);
-	result = imc_scanner->send_message(imc_scanner, connection_id, FALSE, 0,
-									   TNC_IMVID_ANY, msg->get_encoding(msg));	
-	msg->destroy(msg);
 
-	return result;
+	return TNC_RESULT_SUCCESS;
 }
 
 /**
@@ -250,60 +260,108 @@ static TNC_Result send_message(TNC_ConnectionID connection_id)
 TNC_Result TNC_IMC_BeginHandshake(TNC_IMCID imc_id,
 								  TNC_ConnectionID connection_id)
 {
-	if (!imc_scanner)
-	{
-		DBG1(DBG_IMC, "IMC \"%s\" has not been initialized", imc_name);
-		return TNC_RESULT_NOT_INITIALIZED;
-	}
-	return send_message(connection_id);
-}
-
-static TNC_Result receive_message(TNC_IMCID imc_id,
-								  TNC_ConnectionID connection_id,
-								  TNC_UInt32 msg_flags,
-								  chunk_t msg,
-								  TNC_VendorID msg_vid,
-								  TNC_MessageSubtype msg_subtype,
-								  TNC_UInt32 src_imv_id,
-								  TNC_UInt32 dst_imc_id)
-{
-	pa_tnc_msg_t *pa_tnc_msg;
 	imc_state_t *state;
-	TNC_Result result;
-	bool fatal_error;
+	imc_msg_t *out_msg;
+	TNC_Result result = TNC_RESULT_SUCCESS;
 
 	if (!imc_scanner)
 	{
 		DBG1(DBG_IMC, "IMC \"%s\" has not been initialized", imc_name);
 		return TNC_RESULT_NOT_INITIALIZED;
 	}
-
-	/* get current IMC state */
 	if (!imc_scanner->get_state(imc_scanner, connection_id, &state))
 	{
 		return TNC_RESULT_FATAL;
 	}
+	if (lib->settings->get_bool(lib->settings,
+								"libimcv.plugins.imc-scanner.push_info", TRUE))
+	{
+		out_msg = imc_msg_create(imc_scanner, state, connection_id, imc_id,
+								 TNC_IMVID_ANY, msg_types[0]);
+		result = add_port_filter(out_msg);
+		if (result == TNC_RESULT_SUCCESS)
+		{
+			/* send PA-TNC message with the excl flag not set */
+			result = out_msg->send(out_msg, FALSE);
+		}
+		out_msg->destroy(out_msg);
+	}
 
-	/* parse received PA-TNC message and automatically handle any errors */ 
-	result = imc_scanner->receive_message(imc_scanner, state, msg, msg_vid,
-							msg_subtype, src_imv_id, dst_imc_id, &pa_tnc_msg);
+	return result;
+}
 
-	/* no parsed PA-TNC attributes available if an error occurred */
-	if (!pa_tnc_msg)
+static TNC_Result receive_message(imc_msg_t *in_msg)
+{
+	imc_msg_t *out_msg;
+	enumerator_t *enumerator;
+	pa_tnc_attr_t *attr;
+	pen_type_t attr_type;
+	TNC_Result result = TNC_RESULT_SUCCESS;
+	bool fatal_error = FALSE;
+
+	/* parse received PA-TNC message and handle local and remote errors */
+	result = in_msg->receive(in_msg, &fatal_error);
+	if (result != TNC_RESULT_SUCCESS)
 	{
 		return result;
 	}
+	out_msg = imc_msg_create_as_reply(in_msg);
 
-	/* preprocess any IETF standard error attributes */
-	fatal_error = pa_tnc_msg->process_ietf_std_errors(pa_tnc_msg);
-	pa_tnc_msg->destroy(pa_tnc_msg);
+	/* analyze PA-TNC attributes */
+	enumerator = in_msg->create_attribute_enumerator(in_msg);
+	while (enumerator->enumerate(enumerator, &attr))
+	{
+		attr_type = attr->get_type(attr);
 
-	/* if no error occurred then always return the same response */
-	return fatal_error ? TNC_RESULT_FATAL : send_message(connection_id);
+		if (attr_type.vendor_id != PEN_IETF)
+		{
+			continue;
+		}
+		if (attr_type.type == IETF_ATTR_ATTRIBUTE_REQUEST)
+		{
+			ietf_attr_attr_request_t *attr_cast;
+			pen_type_t *entry;
+			enumerator_t *e;
+
+			attr_cast = (ietf_attr_attr_request_t*)attr;
+
+			e = attr_cast->create_enumerator(attr_cast);
+			while (e->enumerate(e, &entry))
+			{
+				if (entry->vendor_id != PEN_IETF)
+				{
+					continue;
+				}
+				switch (entry->type)
+				{
+					case IETF_ATTR_PORT_FILTER:
+						result = add_port_filter(out_msg);
+						break;
+					default:
+						break;
+				}
+			}
+			e->destroy(e);
+		}
+	}
+	enumerator->destroy(enumerator);
+
+	if (fatal_error)
+	{
+		result = TNC_RESULT_FATAL;
+	}
+	else if (result == TNC_RESULT_SUCCESS)
+	{
+		result = out_msg->send(out_msg, TRUE);
+	}
+	out_msg->destroy(out_msg);
+
+	return result;
 }
 
 /**
  * see section 3.8.4 of TCG TNC IF-IMC Specification 1.3
+
  */
 TNC_Result TNC_IMC_ReceiveMessage(TNC_IMCID imc_id,
 								  TNC_ConnectionID connection_id,
@@ -311,14 +369,26 @@ TNC_Result TNC_IMC_ReceiveMessage(TNC_IMCID imc_id,
 								  TNC_UInt32 msg_len,
 								  TNC_MessageType msg_type)
 {
-	TNC_VendorID msg_vid;
-	TNC_MessageSubtype msg_subtype;
+	imc_state_t *state;
+	imc_msg_t *in_msg;
+	TNC_Result result;
 
-	msg_vid = msg_type >> 8;
-	msg_subtype = msg_type & TNC_SUBTYPE_ANY;
+	if (!imc_scanner)
+	{
+		DBG1(DBG_IMC, "IMC \"%s\" has not been initialized", imc_name);
+		return TNC_RESULT_NOT_INITIALIZED;
+	}
+	if (!imc_scanner->get_state(imc_scanner, connection_id, &state))
+	{
+		return TNC_RESULT_FATAL;
+	}
 
-	return receive_message(imc_id, connection_id, 0, chunk_create(msg, msg_len),
-						   msg_vid,	msg_subtype, 0, TNC_IMCID_ANY);
+	in_msg = imc_msg_create_from_data(imc_scanner, state, connection_id,
+									  msg_type, chunk_create(msg, msg_len));
+	result = receive_message(in_msg);
+	in_msg->destroy(in_msg);
+
+	return result;
 }
 
 /**
@@ -334,9 +404,26 @@ TNC_Result TNC_IMC_ReceiveMessageLong(TNC_IMCID imc_id,
 									  TNC_UInt32 src_imv_id,
 									  TNC_UInt32 dst_imc_id)
 {
-	return receive_message(imc_id, connection_id, msg_flags,
-						   chunk_create(msg, msg_len), msg_vid, msg_subtype,
-						   src_imv_id, dst_imc_id);
+	imc_state_t *state;
+	imc_msg_t *in_msg;
+	TNC_Result result;
+
+	if (!imc_scanner)
+	{
+		DBG1(DBG_IMC, "IMC \"%s\" has not been initialized", imc_name);
+		return TNC_RESULT_NOT_INITIALIZED;
+	}
+	if (!imc_scanner->get_state(imc_scanner, connection_id, &state))
+	{
+		return TNC_RESULT_FATAL;
+	}
+	in_msg = imc_msg_create_from_long_data(imc_scanner, state, connection_id,
+								src_imv_id, dst_imc_id, msg_vid, msg_subtype,
+								chunk_create(msg, msg_len));
+	result = receive_message(in_msg);
+	in_msg->destroy(in_msg);
+
+	return result;
 }
 
 /**
