@@ -25,8 +25,7 @@
 
 #include <stdarg.h>
 
-#include <utils.h>
-#include <chunk.h>
+#include <library.h>
 
 /**
  * Definition of some primitive ASN1 types
@@ -36,8 +35,8 @@ typedef enum {
 	ASN1_BOOLEAN =			0x01,
 	ASN1_INTEGER =			0x02,
 	ASN1_BIT_STRING =		0x03,
-	ASN1_OCTET_STRING = 	0x04,
-	ASN1_NULL = 			0x05,
+	ASN1_OCTET_STRING =		0x04,
+	ASN1_NULL =				0x05,
 	ASN1_OID =				0x06,
 	ASN1_ENUMERATED =		0x0A,
 	ASN1_UTF8STRING =		0x0C,
@@ -49,7 +48,7 @@ typedef enum {
 	ASN1_UTCTIME =			0x17,
 	ASN1_GENERALIZEDTIME =	0x18,
 	ASN1_GRAPHICSTRING =	0x19,
-	ASN1_VISIBLESTRING = 	0x1A,
+	ASN1_VISIBLESTRING =	0x1A,
 	ASN1_GENERALSTRING =	0x1B,
 	ASN1_UNIVERSALSTRING =	0x1C,
 	ASN1_BMPSTRING =		0x1E,
@@ -76,7 +75,7 @@ typedef enum {
 	ASN1_CONTEXT_C_4 =		0xA4,
 	ASN1_CONTEXT_C_5 =		0xA5,
 
-	ASN1_INVALID = 			0x100,
+	ASN1_INVALID =			0x100,
 } asn1_t;
 
 #define ASN1_INVALID_LENGTH	0xffffffff
@@ -115,6 +114,22 @@ int asn1_known_oid(chunk_t object);
  */
 chunk_t asn1_build_known_oid(int n);
 
+/**
+ * Convert human readable OID to ASN.1 DER encoding, without OID header.
+ *
+ * @param str		OID string (e.g. 1.2.345.67.8)
+ * @return			allocated ASN.1 encoded OID, chunk_empty on error
+ */
+chunk_t asn1_oid_from_string(char *str);
+
+/**
+ * Convert a DER encoded ASN.1 OID to a human readable string.
+ *
+ * @param oid		DER encoded OID, without header
+ * @return			human readable OID string, allocated, NULL on error
+ */
+char* asn1_oid_to_string(chunk_t oid);
+
 /**
  * Returns the length of an ASN.1 object
  * The blob pointer is advanced past the tag length fields
@@ -155,6 +170,15 @@ int asn1_parse_algorithmIdentifier(chunk_t blob, int level0, chunk_t *params);
 bool asn1_parse_simple_object(chunk_t *object, asn1_t type, u_int level0,
 							  const char* name);
 
+/**
+ * Converts an ASN.1 INTEGER object to an u_int64_t. If the INTEGER is longer
+ * than 8 bytes only the 8 LSBs are returned.
+ *
+ * @param blob		body of an ASN.1 coded integer object
+ * @return			converted integer
+ */
+u_int64_t asn1_parse_integer_uint64(chunk_t blob);
+
 /**
  * Print the value of an ASN.1 simple object
  *
@@ -176,6 +200,8 @@ time_t asn1_to_time(const chunk_t *utctime, asn1_t type);
 /**
  * Converts time_t to an ASN.1 UTCTIME or GENERALIZEDTIME string
  *
+ * @note The type is automatically changed to GENERALIZEDTIME if needed
+ *
  * @param time		time_t in UTC
  * @param type		ASN1_UTCTIME or ASN1_GENERALIZEDTIME
  * @return			body of an ASN.1 code time object
