@@ -40,8 +40,13 @@ struct random_plugin_t {
 };
 
 /**
- * Create a random_plugin instance.
+ * Get the /dev/random file descriptor
  */
-plugin_t *plugin_create();
+int random_plugin_get_dev_random();
+
+/**
+ * Get the /dev/urandom file descriptor
+ */
+int random_plugin_get_dev_urandom();
 
 #endif /** RANDOM_PLUGIN_H_ @}*/
