@@ -13,8 +13,8 @@
  * for more details.
  */
 
-#include <dispatcher.h>
-#include <debug.h>
+#include <fast_dispatcher.h>
+#include <utils/debug.h>
 #include <stdio.h>
 
 #include "manager.h"
@@ -27,7 +27,7 @@
 
 int main (int arc, char *argv[])
 {
-	dispatcher_t *dispatcher;
+	fast_dispatcher_t *dispatcher;
 	storage_t *storage;
 	char *socket;
 	char *database;
@@ -35,7 +35,7 @@ int main (int arc, char *argv[])
 	int threads, timeout;
 
 	library_init(NULL);
-	if (!lib->plugins->load(lib->plugins, NULL,
+	if (!lib->plugins->load(lib->plugins,
 			lib->settings->get_str(lib->settings, "manager.load", PLUGINS)))
 	{
 		return 1;
@@ -48,8 +48,9 @@ int main (int arc, char *argv[])
 	database = lib->settings->get_str(lib->settings, "manager.database", NULL);
 	if (!database)
 	{
-		DBG1("database URI undefined, set manager.database in strongswan.conf");
-		return 1;
+		DBG1(DBG_LIB, "database URI undefined, set manager.database "
+			 "in strongswan.conf");
+		//return 1;
 	}
 
 	storage = storage_create(database);
@@ -58,8 +59,8 @@ int main (int arc, char *argv[])
 		return 1;
 	}
 
-	dispatcher = dispatcher_create(socket, debug, timeout,
-						(context_constructor_t)manager_create, storage);
+	dispatcher = fast_dispatcher_create(socket, debug, timeout,
+						(fast_context_constructor_t)manager_create, storage);
 	dispatcher->add_controller(dispatcher, ikesa_controller_create, NULL);
 	dispatcher->add_controller(dispatcher, gateway_controller_create, NULL);
 	dispatcher->add_controller(dispatcher, auth_controller_create, NULL);
@@ -77,4 +78,3 @@ int main (int arc, char *argv[])
 
 	return 0;
 }
-
